{ *********************************************************************** }
{                                                                         }
{ Faster AnsiString functions for Linux                                   }
{                                                                         }
{ Copyright (c) 2005 Andreas Hausladen                                    }
{                                                                         }
{  This software is provided 'as-is', without any express or              }
{  implied warranty. In no event will the author be held liable           }
{  for any damages arising from the use of this software.                 }
{                                                                         }
{  Permission is granted to anyone to use this software for any           }
{  purpose, including commercial applications, and to alter it            }
{  and redistribute it freely, subject to the following                   }
{  restrictions:                                                          }
{                                                                         }
{    1. The origin of this software must not be misrepresented,           }
{       you must not claim that you wrote the original software.          }
{       If you use this software in a product, an acknowledgment          }
{       in the product documentation would be appreciated but is          }
{       not required.                                                     }
{                                                                         }
{    2. Altered source versions must be plainly marked as such, and       }
{       must not be misrepresented as being the original software.        }
{                                                                         }
{    3. This notice may not be removed or altered from any source         }
{       distribution.                                                     }
{                                                                         }
{ *********************************************************************** }

unit FastLinuxRTL;

interface

uses
  SysUtils;

function FastAnsiCompareText(const S1, S2: AnsiString): Integer;
function FastAnsiUpperCase(const S: AnsiString): AnsiString;
function FastAnsiLowerCase(const S: AnsiString): AnsiString;

function FastWideCompareText(const S1, S2: WideString): Integer;
function FastWideUpperCase(const S: WideString): WideString;
function FastWideLowerCase(const S: WideString): WideString;

function FastInt32ToStr(Value: Longint): AnsiString;
procedure FastMove(const Source; var Dest; Count: Integer);

implementation

{$IFDEF LINUX}

uses
  Libc;

{$ENDIF LINUX}

type
  PStr = ^TStr;
  TStr = packed record
    RefCount: Integer;
    Len: Integer;
  end;

{ Case-mapping "arrays" }
var
  AnsiUpperChars: AnsiString; // 1..255
  AnsiLowerChars: AnsiString; // 1..255
  WideUpperChars: WideString; // 1..65535
  WideLowerChars: WideString; // 1..65535

procedure InitCharArrays;
var
  i: Integer;
begin
  // first initialize the WideChar arrays
  SetLength(WideUpperChars, Word(High(WideChar)) + 1);
  SetLength(WideLowerChars, Word(High(WideChar)) + 1);
  for i := 0 to Word(High(WideChar)) do
    WideUpperChars[i + 1] := WideChar(i);
  for i := 0 to Word(High(WideChar)) do
    WideLowerChars[i + 1] := WideChar(i);
  WideUpperChars := WideUpperCase(WideUpperChars);
  WideLowerChars := WideLowerCase(WideLowerChars);

  // use the widechar array to initialize the AnsiChar arrays
  SetLength(AnsiUpperChars, Byte(High(Char)) + 1);
  SetLength(AnsiLowerChars, Byte(High(Char)) + 1);
  for i := 0 to Byte(High(Char)) do
    AnsiUpperChars[i + 1] := AnsiChar(i);
  for i := 0 to Byte(High(Char)) do
    AnsiLowerChars[i + 1] := AnsiChar(i);
  AnsiUpperChars := FastWideUpperCase(AnsiUpperChars);
  AnsiLowerChars := FastWideLowerCase(AnsiLowerChars);
end;

function FastAnsiCompareText(const S1, S2: AnsiString): Integer;
var
  i: Integer;
  Len, Len1, Len2: Integer;
  ch1, ch2: AnsiChar;
begin
  if Pointer(S1) = Pointer(S2) then
  begin
    Result := 0;
    Exit;
  end;
  if S1 = '' then
  begin
    Result := -1;
    Exit;
  end
  else if S2 = '' then
  begin
    Result := 1;
    Exit;
  end;

  Len1 := PStr(PChar(Pointer(S1)) - SizeOf(TStr))^.Len;
  Len2 := PStr(PChar(Pointer(S2)) - SizeOf(TStr))^.Len;
  if Len2 > Len1 then
    Len := Len2
  else
    Len := Len1;
  if LeadBytes = [] then
  begin
    for i := 0 to Len - 1 do
    begin
      ch1 := S1[i + 1];
      ch2 := S2[i + 1];
      if ch1 <> ch2 then
      begin
        ch1 := AnsiUpperChars[Byte(ch1) + 1];
        ch2 := AnsiUpperChars[Byte(ch2) + 1];

        Result := Ord(ch1) - Ord(ch2);
        if Result <> 0 then
        begin
          if Result < 0 then
            Result := -1
          else
            Result := 1;
          Exit;
        end;
      end;
    end;
    Result := 0;
  end
  else
  begin
    for i := 0 to Len1 - 1 do
    begin
      ch1 := S1[i + 1];
      ch2 := S2[i + 1];
      if ch1 <> ch2 then
      begin
        if (ch1 in LeadBytes) or (ch2 in LeadBytes) then
        begin
          Result := FastWideCompareText(Copy(S1, i, MaxInt), Copy(S2, i, MaxInt));
          Exit;
        end;
        ch1 := AnsiUpperChars[Byte(ch1) + 1];
        ch2 := AnsiUpperChars[Byte(ch2) + 1];

        Result := Ord(ch1) - Ord(ch2);
        if Result <> 0 then
        begin
          if Result < 0 then
            Result := -1
          else
            Result := 1;
          Exit;
        end;
      end;
    end;
    Result := 0;
  end;
end;

function FastAnsiUpperCase(const S: AnsiString): AnsiString;
var
  i: Integer;
begin
  SetLength(Result, Length(S));
  if LeadBytes = [] then
  begin
    for i := 0 to Length(S) - 1 do
      Result[i + 1] := AnsiUpperChars[Byte(S[i + 1]) + 1];
  end
  else
  begin
    for i := 0 to Length(S) - 1 do
    begin
      if S[i + 1] in LeadBytes then
      begin
        Result := FastWideUpperCase(S);
        Exit;
      end;
      Result[i + 1] := AnsiUpperChars[Byte(S[i + 1]) + 1];
    end;
  end;
end;

function FastAnsiLowerCase(const S: AnsiString): AnsiString;
var
  i: Integer;
begin
  SetLength(Result, Length(S));
  if LeadBytes = [] then
  begin
    for i := 0 to Length(S) - 1 do
      Result[i + 1] := AnsiLowerChars[Byte(S[i + 1]) + 1];
  end
  else
  begin
    for i := 0 to Length(S) - 1 do
    begin
      if S[i + 1] in LeadBytes then
      begin
        Result := FastWideLowerCase(S);
        Exit;
      end;
      Result[i + 1] := AnsiLowerChars[Byte(S[i + 1]) + 1];
    end;
  end;
end;

function FastWideCompareText(const S1, S2: WideString): Integer;
const
  EmptyStr = '';
var
  i: Integer;
  Len, Len1, Len2: Integer;
  ch1, ch2: WideChar;
begin
  if Pointer(S1) = Pointer(S2) then
  begin
    Result := 0;
    Exit;
  end;
  Len1 := Length(S1);
  Len2 := Length(S2);
  if Len2 > Len1 then
    Len := Len2
  else
    Len := Len1;
  for i := 0 to Len - 1 do
  begin
    ch1 := S1[i + 1];
    ch2 := S2[i + 1];
    if ch1 <> ch2 then
    begin
      ch1 := WideUpperChars[Word(ch1) + 1];
      ch2 := WideUpperChars[Word(ch2) + 1];

      Result := Integer(Word(ch1)) - Word(ch2);
      if Result <> 0 then
      begin
        if Result < 0 then Result := -1 else Result := 1;
        Exit;
      end;
    end;
  end;
  Result := 0;
end;

function FastWideUpperCase(const S: WideString): WideString;
var
  i: Integer;
begin
  SetLength(Result, Length(S));
  for i := 0 to Length(S) - 1 do
    Result[i + 1] := WideUpperChars[Word(S[i + 1]) + 1];
end;

function FastWideLowerCase(const S: WideString): WideString;
var
  i: Integer;
begin
  SetLength(Result, Length(S));
  for i := 0 to Length(S) - 1 do
    Result[i + 1] := WideLowerChars[Word(S[i + 1]) + 1];
end;

function FastInt32ToStr(Value: Longint): AnsiString;

{  function SmallDivMod(Divident: Integer; var Modulo: Byte; Divisor: Integer): Integer; register;
  asm
    push edx
    xor edx, edx
    div ecx
    pop ecx
    mov [ecx], dl
  end;}

var
  Len: Integer;
  ModValue: Byte;
  Negative: Longint;
  Number: array[0..19] of AnsiChar;
  P, F: PAnsiChar;
begin
  Negative := Value;
  if Negative < 0 then
    Value := -1 * Value;

  Len := 0;
  repeat
    asm
      push eax
      push edx
      push ecx
      mov eax, [Value]
      mov ecx, 10
      xor edx, edx
      div ecx
      mov [ModValue], dl
      mov [Value], eax
      pop ecx
      pop edx
      pop eax
    end;
//    Value := SmallDivMod(Value, ModValue, 10);
    Number[Len] := AnsiChar(ModValue + Ord('0'));
    Inc(Len);
  until Value = 0;

  if Negative >= 0 then
  begin
    SetLength(Result, Len);
    P := Pointer(Result);
  end
  else
  begin
    SetLength(Result, 1 + Len);
    P := Pointer(Result);
    P^ := '-';
    Inc(P);
  end;

  F := @Number[Len - 1];
  while Len > 0 do
  begin
    P^ := F^;
    Dec(F);
    Inc(P);
    Dec(Len);
  end;
end;

(*
Copyright (c) 2004, John O'Harrow (john@almcrest.demon.co.uk)

This software is provided 'as-is', without any express or implied warranty.
In no event will the authors be held liable for any damages arising from the
use of this software.

Permission is granted to anyone to use this software for any purpose, including
commercial applications, and to alter it and redistribute it freely, subject to
the following restrictions:

1. The origin of this software must not be misrepresented; you must not claim
   that you wrote the original software. If you use this software in a product,
   an acknowledgment in the product documentation would be appreciated but is
   not required.

2. Altered source versions must be plainly marked as such, and must not be
   misrepresented as being the original software.

3. This notice may not be removed or altered from any source distribution.

-------------------------------------------------------------------------------

Version: 1.40 - 16-SEP-2004
*)
{Perform Forward Move of 0..36 Bytes}
{On Entry, ECX = Count, EAX = Source+Count, EDX = Dest+Count.  Destroys ECX}
procedure SmallForwardMove_3;
asm
  jmp     dword ptr [@@FwdJumpTable+ecx*4]
  nop {Align Jump Table}
@@FwdJumpTable:
  dd      @@Done {Removes need to test for zero size move}
  dd      @@Fwd01,@@Fwd02,@@Fwd03,@@Fwd04,@@Fwd05,@@Fwd06,@@Fwd07,@@Fwd08
  dd      @@Fwd09,@@Fwd10,@@Fwd11,@@Fwd12,@@Fwd13,@@Fwd14,@@Fwd15,@@Fwd16
  dd      @@Fwd17,@@Fwd18,@@Fwd19,@@Fwd20,@@Fwd21,@@Fwd22,@@Fwd23,@@Fwd24
  dd      @@Fwd25,@@Fwd26,@@Fwd27,@@Fwd28,@@Fwd29,@@Fwd30,@@Fwd31,@@Fwd32
  dd      @@Fwd33,@@Fwd34,@@Fwd35,@@Fwd36
@@Fwd36:
  mov     ecx,[eax-36]
  mov     [edx-36],ecx
@@Fwd32:
  mov     ecx,[eax-32]
  mov     [edx-32],ecx
@@Fwd28:
  mov     ecx,[eax-28]
  mov     [edx-28],ecx
@@Fwd24:
  mov     ecx,[eax-24]
  mov     [edx-24],ecx
@@Fwd20:
  mov     ecx,[eax-20]
  mov     [edx-20],ecx
@@Fwd16:
  mov     ecx,[eax-16]
  mov     [edx-16],ecx
@@Fwd12:
  mov     ecx,[eax-12]
  mov     [edx-12],ecx
@@Fwd08:
  mov     ecx,[eax-8]
  mov     [edx-8],ecx
@@Fwd04:
  mov     ecx,[eax-4]
  mov     [edx-4],ecx
  ret
@@Fwd35:
  mov     ecx,[eax-35]
  mov     [edx-35],ecx
@@Fwd31:
  mov     ecx,[eax-31]
  mov     [edx-31],ecx
@@Fwd27:
  mov     ecx,[eax-27]
  mov     [edx-27],ecx
@@Fwd23:
  mov     ecx,[eax-23]
  mov     [edx-23],ecx
@@Fwd19:
  mov     ecx,[eax-19]
  mov     [edx-19],ecx
@@Fwd15:
  mov     ecx,[eax-15]
  mov     [edx-15],ecx
@@Fwd11:
  mov     ecx,[eax-11]
  mov     [edx-11],ecx
@@Fwd07:
  mov     ecx,[eax-7]
  mov     [edx-7],ecx
  mov     ecx,[eax-4] 
  mov     [edx-4],ecx
  ret
@@Fwd03:
  movzx   ecx, word ptr [eax-3]
  mov     [edx-3],cx
  movzx   ecx, byte ptr [eax-1]
  mov     [edx-1],cl
  ret
@@Fwd34:
  mov     ecx,[eax-34]
  mov     [edx-34],ecx
@@Fwd30:
  mov     ecx,[eax-30]
  mov     [edx-30],ecx
@@Fwd26:
  mov     ecx,[eax-26]
  mov     [edx-26],ecx
@@Fwd22:
  mov     ecx,[eax-22]
  mov     [edx-22],ecx
@@Fwd18:
  mov     ecx,[eax-18]
  mov     [edx-18],ecx
@@Fwd14:
  mov     ecx,[eax-14]
  mov     [edx-14],ecx
@@Fwd10:
  mov     ecx,[eax-10]
  mov     [edx-10],ecx
@@Fwd06:
  mov     ecx,[eax-6]
  mov     [edx-6],ecx
@@Fwd02:
  movzx   ecx, word ptr [eax-2]
  mov     [edx-2],cx
  ret
@@Fwd33:
  mov     ecx,[eax-33]
  mov     [edx-33],ecx
@@Fwd29:
  mov     ecx,[eax-29]
  mov     [edx-29],ecx
@@Fwd25:
  mov     ecx,[eax-25]
  mov     [edx-25],ecx
@@Fwd21:
  mov     ecx,[eax-21]
  mov     [edx-21],ecx
@@Fwd17:
  mov     ecx,[eax-17]
  mov     [edx-17],ecx
@@Fwd13:
  mov     ecx,[eax-13]
  mov     [edx-13],ecx
@@Fwd09:
  mov     ecx,[eax-9]
  mov     [edx-9],ecx
@@Fwd05:
  mov     ecx,[eax-5]
  mov     [edx-5],ecx
@@Fwd01:
  movzx   ecx, byte ptr [eax-1]
  mov     [edx-1],cl
@@Done:
end; {SmallForwardMove}

{-------------------------------------------------------------------------}
{Perform Backward Move of 0..36 Bytes}
{On Entry, ECX = Count, EAX = Source, EDX = Dest.  Destroys ECX}
procedure SmallBackwardMove_3;
asm
  jmp     dword ptr [@@BwdJumpTable+ecx*4]
  nop {Align Jump Table}
@@BwdJumpTable:
  dd      @@Done {Removes need to test for zero size move}
  dd      @@Bwd01,@@Bwd02,@@Bwd03,@@Bwd04,@@Bwd05,@@Bwd06,@@Bwd07,@@Bwd08
  dd      @@Bwd09,@@Bwd10,@@Bwd11,@@Bwd12,@@Bwd13,@@Bwd14,@@Bwd15,@@Bwd16
  dd      @@Bwd17,@@Bwd18,@@Bwd19,@@Bwd20,@@Bwd21,@@Bwd22,@@Bwd23,@@Bwd24
  dd      @@Bwd25,@@Bwd26,@@Bwd27,@@Bwd28,@@Bwd29,@@Bwd30,@@Bwd31,@@Bwd32
  dd      @@Bwd33,@@Bwd34,@@Bwd35,@@Bwd36
@@Bwd36:
  mov     ecx,[eax+32]
  mov     [edx+32],ecx
@@Bwd32:
  mov     ecx,[eax+28]
  mov     [edx+28],ecx
@@Bwd28:
  mov     ecx,[eax+24]
  mov     [edx+24],ecx
@@Bwd24:
  mov     ecx,[eax+20]
  mov     [edx+20],ecx
@@Bwd20:
  mov     ecx,[eax+16]
  mov     [edx+16],ecx
@@Bwd16:
  mov     ecx,[eax+12]
  mov     [edx+12],ecx
@@Bwd12:
  mov     ecx,[eax+8]
  mov     [edx+8],ecx
@@Bwd08:
  mov     ecx,[eax+4]
  mov     [edx+4],ecx
@@Bwd04:
  mov     ecx,[eax]
  mov     [edx],ecx
  ret
@@Bwd35:
  mov     ecx,[eax+31]
  mov     [edx+31],ecx
@@Bwd31:
  mov     ecx,[eax+27]
  mov     [edx+27],ecx
@@Bwd27:
  mov     ecx,[eax+23]
  mov     [edx+23],ecx
@@Bwd23:
  mov     ecx,[eax+19]
  mov     [edx+19],ecx
@@Bwd19:
  mov     ecx,[eax+15]
  mov     [edx+15],ecx
@@Bwd15:
  mov     ecx,[eax+11]
  mov     [edx+11],ecx
@@Bwd11:
  mov     ecx,[eax+7]
  mov     [edx+7],ecx
@@Bwd07:
  mov     ecx,[eax+3]
  mov     [edx+3],ecx
  mov     ecx,[eax]
  mov     [edx],ecx
  ret
@@Bwd03:
  movzx   ecx, word ptr [eax+1]
  mov     [edx+1],cx
  movzx   ecx, byte ptr [eax]
  mov     [edx],cl
  ret
@@Bwd34:
  mov     ecx,[eax+30]
  mov     [edx+30],ecx
@@Bwd30:
  mov     ecx,[eax+26]
  mov     [edx+26],ecx
@@Bwd26:
  mov     ecx,[eax+22]
  mov     [edx+22],ecx
@@Bwd22:
  mov     ecx,[eax+18]
  mov     [edx+18],ecx
@@Bwd18:
  mov     ecx,[eax+14]
  mov     [edx+14],ecx
@@Bwd14:
  mov     ecx,[eax+10]
  mov     [edx+10],ecx
@@Bwd10:
  mov     ecx,[eax+6]
  mov     [edx+6],ecx
@@Bwd06:
  mov     ecx,[eax+2]
  mov     [edx+2],ecx
@@Bwd02:
  movzx   ecx, word ptr [eax]
  mov     [edx],cx
  ret
@@Bwd33:
  mov     ecx,[eax+29]
  mov     [edx+29],ecx
@@Bwd29:
  mov     ecx,[eax+25]
  mov     [edx+25],ecx
@@Bwd25:
  mov     ecx,[eax+21]
  mov     [edx+21],ecx
@@Bwd21:
  mov     ecx,[eax+17]
  mov     [edx+17],ecx
@@Bwd17:
  mov     ecx,[eax+13]
  mov     [edx+13],ecx
@@Bwd13:
  mov     ecx,[eax+9]
  mov     [edx+9],ecx
@@Bwd09:
  mov     ecx,[eax+5]
  mov     [edx+5],ecx
@@Bwd05:
  mov     ecx,[eax+1]
  mov     [edx+1],ecx
@@Bwd01:
  movzx   ecx, byte ptr[eax]
  mov     [edx],cl
@@Done:
end; {SmallBackwardMove}

{-------------------------------------------------------------------------}
{Move ECX Bytes from EAX to EDX, where EAX > EDX and ECX > 36 (SMALLMOVESIZE)}
procedure Forwards_IA32_3;
asm
  push    ebx
  mov     ebx,edx
  fild    qword ptr [eax]
  add     eax,ecx {QWORD Align Writes}
  add     ecx,edx
  add     edx,7
  and     edx,-8
  sub     ecx,edx
  add     edx,ecx {Now QWORD Aligned}
  sub     ecx,16
  neg     ecx
@FwdLoop:
  fild    qword ptr [eax+ecx-16]
  fistp   qword ptr [edx+ecx-16]
  fild    qword ptr [eax+ecx-8]
  fistp   qword ptr [edx+ecx-8]
  add     ecx,16
  jle     @FwdLoop
  fistp   qword ptr [ebx]
  neg     ecx
  add     ecx,16
  pop     ebx
  jmp     SmallForwardMove_3
end; {Forwards_IA32}

{-------------------------------------------------------------------------}
{Move ECX Bytes from EAX to EDX, where EAX < EDX and ECX > 36 (SMALLMOVESIZE)}
procedure Backwards_IA32_3;
asm
  push    ebx
  fild    qword ptr [eax+ecx-8]
  lea     ebx,[edx+ecx] {QWORD Align Writes}
  and     ebx,7
  sub     ecx,ebx
  add     ebx,ecx {Now QWORD Aligned, EBX = Original Length}
  sub     ecx,16
@BwdLoop:
  fild    qword ptr [eax+ecx]
  fild    qword ptr [eax+ecx+8]
  fistp   qword ptr [edx+ecx+8]
  fistp   qword ptr [edx+ecx]
  sub     ecx,16
  jge     @BwdLoop
  fistp   qword ptr [edx+ebx-8]
  add     ecx,16
  pop     ebx
  jmp     SmallBackwardMove_3
end; {Backwards_IA32}

const
  SMALLMOVESIZE = 36;

//procedure MoveJOH_IA32_3(const Source; var Dest; Count : Integer);
procedure FastMove(const Source; var Dest; Count: Integer);
asm
  cmp     ecx,SMALLMOVESIZE
  ja      @Large
  cmp     eax,edx
  lea     eax,[eax+ecx]
  jle     @SmallCheck
@SmallForward:
  add     edx,ecx
  jmp     SmallForwardMove_3
@SmallCheck:
  je      @Done {For Compatibility with Delphi's move for Source = Dest}
  sub     eax,ecx
  jmp     SmallBackwardMove_3
@Large:
  jng     @Done {For Compatibility with Delphi's move for Count < 0}
  cmp     eax,edx
  jg      Forwards_IA32_3
  je      @Done {For Compatibility with Delphi's move for Source = Dest}
  push    eax
  add     eax,ecx
  cmp     eax,edx
  pop     eax
  jg      Backwards_IA32_3 {Source/Dest Overlap}
  jmp     Forwards_IA32_3
@Done:
end; {MoveJOH_IA32}

{------------------------------------------------------------------------------}
{$IFDEF LINUX}

type
  TJumpCode = packed record
    Code: Byte; // $e9  jmp Distance
    Distance: Integer;
  end;

  PRedirectCode = ^TRedirectCode;
  TRedirectCode = packed record
    Code: Word;
    Address: PPointer;
  end;

  TRedirectInfo = record
    Func: Pointer;
    OrgCode: TJumpCode;
  end;

function InjectCode(Addr: Pointer; Code: Pointer; Size: Integer; OldCode: Pointer): Boolean;
var
  SysPageSize, PageSize, AlignedAddr: Integer;
begin
  Result := False;
  SysPageSize := getpagesize;
  PageSize := SysPageSize;
  AlignedAddr := Integer(Addr) and not (PageSize - 1);
  while Integer(Addr) + Size >= AlignedAddr + PageSize do
    Inc(PageSize, SysPageSize);
  if mprotect(Pointer(AlignedAddr), PageSize, PROT_READ or PROT_WRITE or PROT_EXEC) = 0 then
    try
      if OldCode <> nil then
        Move(Addr^, OldCode^, Size);
      Move(Code^, Addr^, Size);
      Result := True;
    except
    end;
end;

function Redirect(Func: Pointer; RedirectFunc: Pointer): TRedirectInfo;
var
  JumpCode: TJumpCode;
begin
  if PRedirectCode(Func)^.Code = $25FF then
    // redirected (in package or shared object)
    Func := PRedirectCode(Func)^.Address^;

  Result.Func := Func;
  Result.OrgCode.Code := 0;
  Result.OrgCode.Distance := 0;

  JumpCode.Code := $e9;
  JumpCode.Distance := Integer(RedirectFunc) -
                       Integer(Func) -
                       SizeOf(TJumpCode);
  if not CompareMem(Func, @JumpCode, SizeOf(TJumpCode)) then
    InjectCode(Func, @JumpCode, SizeOf(TJumpCode), @Result.OrgCode);
end;

procedure RestoreRedirect(const Redir: TRedirectInfo);
var
  Func: Pointer;
begin
  Func := Redir.Func;
  if PRedirectCode(Func)^.Code = $25FF then
    // redirected (in package or shared object)
    Func := PRedirectCode(Func)^.Address^;

  InjectCode(Func, @Redir.OrgCode, SizeOf(TJumpCode), nil);
end;

procedure DynamicLinkFunctions;
var
  A: Char;
begin
  { Call the functions to let the dynamic linker do its job. }
  SysUtils.AnsiCompareText('', '');
  SysUtils.AnsiLowerCase('');
  SysUtils.AnsiUpperCase('');
  SysUtils.WideCompareText('', '');
  SysUtils.IntToStr(1);
  Move(A, A, 0);
end;

var
  RedirAnsiCompareText: TRedirectInfo;
  RedirAnsiUpperCase: TRedirectInfo;
  RedirAnsiLowerCase: TRedirectInfo;
  RedirWideCompareText: TRedirectInfo;
  RedirWideUpperCase: TRedirectInfo;
  RedirWideLowerCase: TRedirectInfo;
  RedirInt32ToStr: TRedirectInfo;
  RedirMove: TRedirectInfo;

{$ENDIF LINUX}

initialization
  InitCharArrays; // execute before redirect
{$IFDEF LINUX}
  DynamicLinkFunctions;

  { redirect to faster functions }
  RedirAnsiCompareText := Redirect(@AnsiCompareText, @FastAnsiCompareText);
  RedirAnsiUpperCase := Redirect(@AnsiUpperCase, @FastAnsiUpperCase);
  RedirAnsiLowerCase := Redirect(@AnsiLowerCase, @FastAnsiLowerCase);

  RedirWideCompareText := Redirect(@WideCompareText, @FastWideCompareText);
  RedirWideUpperCase := Redirect(@WideUpperCase, @FastWideUpperCase);
  RedirWideLowerCase := Redirect(@WideLowerCase, @FastWideLowerCase);

//  RedirInt32ToStr := Redirect(@IntToStr, @FastInt32ToStr); // @IntToStr returns @IntToStr(Integer)
  RedirMove := Redirect(@Move, @FastMove);

finalization
  RestoreRedirect(RedirAnsiCompareText);
  RestoreRedirect(RedirAnsiUpperCase);
  RestoreRedirect(RedirAnsiLowerCase);

  RestoreRedirect(RedirWideCompareText);
  RestoreRedirect(RedirWideUpperCase);
  RestoreRedirect(RedirWideLowerCase);

//  RestoreRedirect(RedirInt32ToStr);
  RestoreRedirect(RedirMove);
{$ENDIF LINUX}

end.

